"""
TODO
"""
from pathlib import Path

from cppython_core.schema import CPPythonModel

from cppython.utility import read_model_json, write_model_json


class TestBuilder:
    """
    TODO
    """

    class TestModel(CPPythonModel):
        """
        TODO
        """

        test_path: Path
        test_int: int

    def test_model_read_write(self, tmpdir):
        """
        TODO
        """

        test_model = TestBuilder.TestModel(test_path=Path(), test_int=3)

        temporary_directory = Path(tmpdir)

        json_path = temporary_directory / "test.json"

        write_model_json(json_path, test_model)
        output = read_model_json(json_path, TestBuilder.TestModel)

        assert test_model == output



"""
Test the functions related to the internal interface implementation and the 'Interface' interface itself
"""

from pathlib import Path

from cppython_core.schema import (
    PEP621,
    ConfigurePreset,
    CPPythonData,
    Generator,
    GeneratorConfiguration,
    GeneratorData,
    PyProject,
    TargetEnum,
    ToolData,
)
from pytest_mock import MockerFixture

from cppython.project import Project, ProjectBuilder, ProjectConfiguration
from cppython.utility import read_json, write_json

default_pep621 = PEP621(name="test_name", version="1.0")
default_cppython_data = CPPythonData(**{"target": TargetEnum.EXE})
default_tool_data = ToolData(**{"cppython": default_cppython_data})
default_pyproject = PyProject(**{"project": default_pep621, "tool": default_tool_data})


class MockGeneratorData(GeneratorData):
    """
    TODO
    """

    check: bool


class ExtendedCPPython(CPPythonData):
    """
    TODO
    """

    mock: MockGeneratorData


class TestProject:
    """
    TODO
    """

    def test_construction(self, mocker: MockerFixture):
        """
        Makes sure the project can be created from the default PyProject data
        """

        interface_mock = mocker.MagicMock()
        configuration = ProjectConfiguration(root_path=Path(), version="1.0.0")
        Project(configuration, interface_mock, default_pyproject.dict(by_alias=True))


class TestBuilder:
    """
    TODO
    """

    def test_plugin_gather(self):
        """
        TODO
        """

        configuration = ProjectConfiguration(root_path=Path(), version="1.0.0")
        builder = ProjectBuilder(configuration)
        plugins = builder.gather_plugins(Generator)

        assert len(plugins) == 0

    def test_generator_data_construction(self, mocker: MockerFixture):
        """
        TODO
        """

        configuration = ProjectConfiguration(root_path=Path(), version="1.0.0")
        builder = ProjectBuilder(configuration)
        model_type = builder.generate_model([])

        assert model_type.__base__ == PyProject

        generator_type = mocker.Mock()
        generator_type.name.return_value = "mock"
        generator_type.data_type.return_value = MockGeneratorData

        model_type = builder.generate_model([generator_type])

        project_data = default_pyproject.dict(by_alias=True)

        mock_data = MockGeneratorData(check=True)
        project_data["tool"]["cppython"]["mock"] = mock_data.dict(by_alias=True)
        result = model_type(**project_data)

        assert result.tool is not None
        assert result.tool.cppython is not None
        assert result.tool.cppython.mock is not None
        assert result.tool.cppython.mock.check

    def test_generator_creation(self, mocker: MockerFixture):
        """
        TODO
        """

        configuration = ProjectConfiguration(root_path=Path(), version="1.0.0")
        builder = ProjectBuilder(configuration)

        generator_configuration = GeneratorConfiguration(root_path=configuration.root_path)
        generators = builder.create_generators([], generator_configuration, default_pep621, default_cppython_data)

        assert not generators

        generator_type = mocker.Mock()
        generator_type.name.return_value = "mock"
        generator_type.data_type.return_value = MockGeneratorData

        mock_data = MockGeneratorData(check=True)
        extended_cppython_dict = default_cppython_data.dict(exclude_defaults=True)
        extended_cppython_dict["mock"] = mock_data
        extended_cppython = ExtendedCPPython(**extended_cppython_dict)

        generators = builder.create_generators(
            [generator_type],
            generator_configuration,
            default_pep621,
            extended_cppython,
        )

        assert len(generators) == 1

    def test_presets(self, tmpdir):
        """
        TODO
        """

        temporary_directory = Path(tmpdir)
        configuration = ProjectConfiguration(root_path=temporary_directory, version="1.0.0")
        builder = ProjectBuilder(configuration)

        input_toolchain = temporary_directory / "input.cmake"

        with open(input_toolchain, "w", encoding="utf8") as file:
            file.write("")

        configure_preset = ConfigurePreset(name="test_preset", toolchainFile=str(input_toolchain))

        generator_output = [("test", configure_preset)]
        builder.write_presets(temporary_directory, generator_output)

        cppython_tool = temporary_directory / "cppython"
        assert cppython_tool.exists()

        cppython_file = cppython_tool / "cppython.json"
        assert cppython_file.exists()

        test_tool = cppython_tool / "test"
        assert test_tool.exists()

        test_file = test_tool / "test.json"
        assert test_file.exists()

    def test_root_unmodified(self, tmpdir):
        """
        TODO
        """

        temporary_directory = Path(tmpdir)
        configuration = ProjectConfiguration(root_path=temporary_directory, version="1.0.0")
        builder = ProjectBuilder(configuration)

        # TODO: Translate into reuseable testing data
        output = {
            "version": 4,
            "cmakeMinimumRequired": {"major": 3, "minor": 23, "patch": 1},
            "include": ["should/be/replaced/cppython.json"],
            "configurePresets": [
                {
                    "name": "default",
                    "inherits": ["cppython"],
                    "hidden": True,
                    "description": "Tests that generator isn't removed",
                    "generator": "Should exist",
                },
            ],
        }

        input_preset = temporary_directory / "CMakePresets.json"
        write_json(input_preset, output)

        builder.write_root_presets(temporary_directory / "test_location")

        data = read_json(input_preset)

        # TODO: Assert the differences affect nothing but what is written by the builder



"""
TODO
"""

import json
from pathlib import Path
from typing import Any, Type

from cppython_core.schema import ModelT
from pydantic import BaseModel


def read_model_json(path: Path, model: Type[ModelT]) -> ModelT:
    """
    Reading routine. Only keeps Model data
    """

    return model.parse_file(path=path)


def read_json(path: Path) -> Any:
    """
    Reading routine
    """

    with open(path, "r", encoding="utf-8") as file:
        return json.load(file)


def write_model_json(path: Path, model: BaseModel) -> None:
    """
    Writing routine. Only writes model data
    """

    serialized = json.loads(model.json(exclude_none=True))
    with open(path, "w", encoding="utf8") as file:
        json.dump(serialized, file, ensure_ascii=False, indent=4)


def write_json(path: Path, data: Any) -> None:
    """
    Writing routine
    """

    with open(path, "w", encoding="utf-8") as file:
        json.dump(data, file, ensure_ascii=False, indent=4)



"""
TODO
"""

from __future__ import annotations  # Required for self-referenced pydantic types

from abc import abstractmethod
from pathlib import Path
from typing import Any, Optional

from cppython_core.schema import ConfigurePreset, CPPythonModel, Preset
from pydantic import Extra, Field, validator


class BuildPreset(Preset):
    """
    Partial Build Preset specification
    """

    configurePreset: Optional[str] = Field(default=None)
    inheritConfigureEnvironment: Optional[bool] = Field(default=None)


class TestPreset(Preset):
    """
    Partial Test Preset specification
    """

    configurePreset: Optional[str] = Field(default=None)
    inheritConfigureEnvironment: Optional[bool] = Field(default=None)


class CMakeVersion(CPPythonModel, extra=Extra.forbid):
    """
    The version specification for CMake
    """

    major: int = Field(default=3)
    minor: int = Field(default=23)
    patch: int = Field(default=1)


class CMakePresets(CPPythonModel, extra=Extra.forbid):
    """
    The schema for the CMakePresets and CMakeUserPresets files
    """

    version: int = Field(default=4, const=True)
    cmakeMinimumRequired: CMakeVersion = Field(default=CMakeVersion())  # TODO: 'version' compatibility validation
    include: Optional[list[str]] = Field(default=None)
    vendor: Optional[Any] = Field(default=None)
    configurePresets: Optional[list[ConfigurePreset]] = Field(default=None)
    buildPresets: Optional[list[BuildPreset]] = Field(default=None)
    testPresets: Optional[list[TestPreset]] = Field(default=None)

    @validator("include")
    def validate_path(cls, values):  # pylint: disable=E0213
        """
        TODO
        """
        if values is not None:
            output = []
            for value in values:
                output.append(Path(value).as_posix())
            return output

        return None


class ProjectConfiguration(CPPythonModel):
    """
    TODO
    """

    root_path: Path  # The path where the pyproject.toml lives
    version: str  # The version number a 'dynamic' project version will resolve to
    verbosity: int = Field(default=0)

    @validator("verbosity")
    def min_max(cls, value):  # pylint: disable=E0213
        """
        TODO
        """
        return min(max(value, 0), 2)


class API:
    """
    Project API
    """

    @abstractmethod
    def install(self) -> None:
        """
        TODO
        """
        raise NotImplementedError()

    @abstractmethod
    def update(self) -> None:
        """
        TODO
        """

        raise NotImplementedError()



"""
TODO
"""


from abc import ABC, abstractmethod
from pathlib import Path

from packaging.version import Version


class VCS(ABC):
    """
    Base class for version control systems
    """

    subclasses = []

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.subclasses.append(cls)

    @abstractmethod
    def is_repository(self, path: Path) -> bool:
        """
        TODO
        """
        raise NotImplementedError()

    @abstractmethod
    def extract_version(self, path: Path) -> Version:
        """
        TODO
        """
        raise NotImplementedError()



"""
TODO
"""


from pathlib import Path

from cppython.console.vcs.git import Git


class TestGit:
    """
    TODO
    """

    def test_version(self):
        """
        TODO
        """

        directory = Path()

        git = Git()

        result = git.extract_version(directory)

        assert result != ""



"""
TODO
"""

from pathlib import Path

from dulwich.porcelain import tag_list
from dulwich.repo import Repo
from packaging.version import Version

from cppython.console.vcs.base import VCS


class Git(VCS):
    """
    Git implementation hooks
    """

    def is_repository(self, path: Path) -> bool:
        """
        TODO
        """

        try:
            Repo(str(path))
            return True

        except Exception:
            return False

    def extract_version(self, path: Path) -> Version:
        """
        TODO
        """

        repo = Repo(str(path))
        tags = tag_list(repo)

        try:
            tag = tags[-1].decode("utf-8")
        except Exception:
            tag = "v0.1.0"
        return Version(tag)



"""
The central delegation of the CPPython project
"""

import logging
from importlib import metadata
from pathlib import Path
from typing import Any, Type, TypeVar

from cppython_core.schema import (
    PEP621,
    CPPythonData,
    CPPythonDataT,
    Generator,
    GeneratorConfiguration,
    Interface,
    Plugin,
    ProjectDataT,
    PyProject,
    ToolData,
)
from cppython_core.utility import cppython_logger
from pydantic import create_model

from cppython.schema import API, CMakePresets, ConfigurePreset, ProjectConfiguration
from cppython.utility import read_json, write_json, write_model_json


class ProjectBuilder:
    """
    TODO
    """

    def __init__(self, configuration: ProjectConfiguration) -> None:
        self.configuration = configuration

    DerivedPlugin = TypeVar("DerivedPlugin", bound=Plugin)

    def gather_plugins(self, plugin_type: Type[DerivedPlugin]) -> list[Type[DerivedPlugin]]:
        """
        TODO
        """
        plugins = []
        entry_points = metadata.entry_points(group=f"cppython.{plugin_type.group()}")

        for entry_point in entry_points:
            loaded_plugin_type = entry_point.load()
            if issubclass(loaded_plugin_type, plugin_type) & (loaded_plugin_type is not plugin_type):
                plugins.append(loaded_plugin_type)

        return plugins

    def generate_model(self, plugins: list[Type[Generator]]) -> Type[PyProject]:
        """
        TODO: Proper return type hint
        """
        plugin_fields = {}
        for plugin_type in plugins:
            plugin_fields[plugin_type.name()] = (plugin_type.data_type(), ...)

        extended_cppython_type = create_model(
            "ExtendedCPPythonData",
            **plugin_fields,
            __base__=CPPythonData,
        )

        extended_tool_type = create_model(
            "ExtendedToolData",
            cppython=(extended_cppython_type, ...),
            __base__=ToolData,
        )

        return create_model(
            "ExtendedPyProject",
            tool=(extended_tool_type, ...),
            __base__=PyProject,
        )

    def create_generators(
        self,
        plugins: list[Type[Generator]],
        configuration: GeneratorConfiguration,
        project: PEP621,
        cppython: CPPythonData,
    ) -> list[Generator]:
        """
        TODO
        """
        _generators = []
        for plugin_type in plugins:
            name = plugin_type.name()
            _generators.append(plugin_type(configuration, project, cppython, getattr(cppython, name)))

        return _generators

    def generate_modified_project(self, original: ProjectDataT) -> ProjectDataT:
        """
        Applies dynamic behaviors of the settings to itself
        Returns a copy of the original with dynamic modifications
        """
        modified = original.copy(deep=True)

        # Update the dynamic version

        modified.version = self.configuration.version

        return modified

    def generate_modified_cppython(self, original: CPPythonDataT) -> CPPythonDataT:
        """
        Applies dynamic behaviors of the settings to itself
        Returns a copy of the original with dynamic modifications
        """
        modified = original.copy(deep=True)

        # Add the pyproject.toml location to all relative paths

        if not modified.install_path.is_absolute():
            modified.install_path = self.configuration.root_path.absolute() / modified.install_path

        if not modified.tool_path.is_absolute():
            modified.tool_path = self.configuration.root_path.absolute() / modified.tool_path

        if not modified.build_path.is_absolute():
            modified.build_path = self.configuration.root_path.absolute() / modified.build_path

        return modified

    def write_presets(self, path: Path, generator_output: list[tuple[str, ConfigurePreset]]) -> Path:
        """
        Write the cppython presets.
        Returns the
        """

        path.mkdir(parents=True, exist_ok=True)

        def write_generator_presets(path: Path, generator_name: str, configure_preset: ConfigurePreset) -> Path:
            """
            Write a generator preset.
            @returns - The written json file
            """
            generator_tool_path = path / generator_name
            generator_tool_path.mkdir(parents=True, exist_ok=True)

            configure_preset.hidden = True
            presets = CMakePresets(configurePresets=[configure_preset])

            json_path = generator_tool_path / f"{generator_name}.json"

            write_model_json(json_path, presets)

            return json_path

        names = []
        includes = []

        path = path / "cppython"

        for generator_name, configure_preset in generator_output:

            preset_file = write_generator_presets(path, generator_name, configure_preset)

            relative_file = preset_file.relative_to(path)

            names.append(generator_name)
            includes.append(str(relative_file))

        configure_preset = ConfigurePreset(name="cppython", hidden=True, inherits=names)
        presets = CMakePresets(configurePresets=[configure_preset], include=includes)

        json_path = path / "cppython.json"

        write_model_json(json_path, presets)
        return json_path

    def write_root_presets(self, path: Path):
        """
        Read the top level json file and replace the include reference.
        Receives a relative path to the tool cmake json file
        """

        root_preset_path = self.configuration.root_path / "CMakePresets.json"

        root_preset = read_json(root_preset_path)
        root_model = CMakePresets.parse_obj(root_preset)

        if root_model.include is not None:
            for index, include_path in enumerate(root_model.include):
                if Path(include_path).name == "cppython.json":
                    root_model.include[index] = "build/" + path.as_posix()

            # 'dict.update' wont apply to nested types, manual replacement
            root_preset["include"] = root_model.include

            write_json(root_preset_path, root_preset)


class Project(API):
    """
    The object constructed at each entry_point
    """

    def __init__(
        self, configuration: ProjectConfiguration, interface: Interface, pyproject_data: dict[str, Any]
    ) -> None:

        self._enabled = False
        self._configuration = configuration

        levels = [logging.WARNING, logging.INFO, logging.DEBUG]

        # Add default output stream
        console_handler = logging.StreamHandler()
        cppython_logger.addHandler(console_handler)
        cppython_logger.setLevel(levels[configuration.verbosity])

        cppython_logger.info("Initializing project")

        self._builder = ProjectBuilder(self.configuration)
        plugins = self._builder.gather_plugins(Generator)

        if not plugins:
            cppython_logger.error("No generator plugin was found")
            return

        for plugin in plugins:
            cppython_logger.warning(f"Generator plugin found: {plugin.name()}")

        extended_pyproject_type = self._builder.generate_model(plugins)
        pyproject = extended_pyproject_type(**pyproject_data)

        if pyproject is None:
            cppython_logger.error("Data is not defined")
            return

        if pyproject.tool is None:
            cppython_logger.error("Table [tool] is not defined")
            return

        if pyproject.tool.cppython is None:
            cppython_logger.error("Table [tool.cppython] is not defined")
            return

        self._enabled = True

        self._project = pyproject.project

        self._modified_project_data = self._builder.generate_modified_project(pyproject.project)
        self._modified_cppython_data = self._builder.generate_modified_cppython(pyproject.tool.cppython)

        self._interface = interface

        generator_configuration = GeneratorConfiguration(root_path=self.configuration.root_path)
        self._generators = self._builder.create_generators(
            plugins, generator_configuration, self.project, self.cppython
        )

        cppython_logger.info("Initialized project successfully")

    @property
    def enabled(self) -> bool:
        """
        TODO
        """
        return self._enabled

    @property
    def configuration(self) -> ProjectConfiguration:
        """
        TODO
        """
        return self._configuration

    @property
    def project(self) -> PEP621:
        """
        The resolved pyproject project table
        """
        return self._modified_project_data

    @property
    def cppython(self):
        """
        The resolved CPPython data
        """
        return self._modified_cppython_data

    def download_generator_tools(self) -> None:
        """
        Download the generator tooling if required
        """
        if not self._enabled:
            cppython_logger.info("Skipping 'download_generator_tools' because the project is not enabled")
            return

        base_path = self.cppython.install_path

        for generator in self._generators:

            path = base_path / generator.name()

            path.mkdir(parents=True, exist_ok=True)

            if not generator.generator_downloaded(path):
                cppython_logger.warning(f"Downloading the {generator.name()} requirements to {path}")

                # TODO: Make async with progress bar
                generator.download_generator(path)
                cppython_logger.warning("Download complete")
            else:
                cppython_logger.info(f"The {generator.name()} generator is already downloaded")

    def update_generator_tools(self) -> None:
        """
        Update the generator tooling if available
        """
        if not self._enabled:
            cppython_logger.info("Skipping 'update_generator_tools' because the project is not enabled")
            return

        self.download_generator_tools()

        base_path = self.cppython.install_path

        for generator in self._generators:

            path = base_path / generator.name()

            generator.update_generator(path)

    # API Contract
    def install(self) -> None:
        """
        TODO
        """
        if not self._enabled:
            cppython_logger.info("Skipping install because the project is not enabled")
            return

        cppython_logger.info("Installing tools")
        self.download_generator_tools()

        cppython_logger.info("Installing project")
        preset_path = self.cppython.build_path

        generator_output = []

        # TODO: Async
        for generator in self._generators:
            cppython_logger.info(f"Installing {generator.name()} generator")

            try:
                generator.install()
                config_preset = generator.generate_cmake_config()
                generator_output.append((generator.name(), config_preset))
            except Exception as exception:
                cppython_logger.error(f"Generator {generator.name()} failed to install")
                raise exception

        project_presets = self._builder.write_presets(preset_path, generator_output)
        self._builder.write_root_presets(project_presets.relative_to(preset_path))

    def update(self) -> None:
        """
        TODO
        """
        if not self._enabled:
            cppython_logger.info("Skipping update because the project is not enabled")
            return

        cppython_logger.info("Updating tools")
        self.update_generator_tools()

        cppython_logger.info("Updating project")

        preset_path = self.cppython.build_path

        generator_output = []

        # TODO: Async
        for generator in self._generators:
            cppython_logger.info(f"Updating {generator.name()} generator")

            try:
                generator.update()
                config_preset = generator.generate_cmake_config()
                generator_output.append((generator.name(), config_preset))
            except Exception as exception:
                cppython_logger.error(f"Generator {generator.name()} failed to update")
                raise exception

        project_presets = self._builder.write_presets(preset_path, generator_output)
        self._builder.write_root_presets(project_presets.relative_to(preset_path))
